@@ -1,21 +1,22 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from itertools import groupby
-
 import stdnum.eu.vat as vat
 import stdnum.exceptions
-from sql import Null
+from sql import Null, Column
 from sql.functions import CharLength
 
-from trytond.model import ModelView, ModelSQL, fields, Unique
+from trytond.model import (ModelView, ModelSQL, MultiValueMixin, ValueMixin,
+    fields, Unique, sequence_ordered)
 from trytond.wizard import Wizard, StateTransition, StateView, Button
-from trytond.pyson import Eval
+from trytond.pyson import Eval, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond import backend
+from trytond.tools.multivalue import migrate_property
 
-__all__ = ['Party', 'PartyCategory', 'PartyIdentifier',
-    'CheckVIESResult', 'CheckVIES']
+__all__ = ['Party', 'PartyLang', 'PartyCategory', 'PartyIdentifier',
+    'CheckVIESResult', 'CheckVIES',
+    'PartyReplace', 'PartyReplaceAsk']
 
 VAT_COUNTRIES = [('', '')]
 STATES = {
@@ -24,31 +25,49 @@ STATES = {
 DEPENDS = ['active']
 
 
-class Party(ModelSQL, ModelView):
+class Party(ModelSQL, ModelView, MultiValueMixin):
     "Party"
     __name__ = 'party.party'
 
-    name = fields.Char('Name', select=True, states=STATES, depends=DEPENDS)
+    name = fields.Char('Name', select=True, states=STATES, depends=DEPENDS,
+        help="The main identifier of the party.")
     code = fields.Char('Code', required=True, select=True,
         states={
             'readonly': Eval('code_readonly', True),
             },
-        depends=['code_readonly'])
+        depends=['code_readonly'],
+        help="The unique identifier of the party.")
     code_readonly = fields.Function(fields.Boolean('Code Readonly'),
         'get_code_readonly')
-    lang = fields.Property(fields.Many2One("ir.lang", 'Language',
-            states=STATES, depends=DEPENDS))
+    lang = fields.MultiValue(
+        fields.Many2One('ir.lang', "Language", states=STATES, depends=DEPENDS,
+            help="Used to translate communications with the party."))
+    langs = fields.One2Many(
+        'party.party.lang', 'party', "Languages")
     identifiers = fields.One2Many('party.identifier', 'party', 'Identifiers',
-        states=STATES, depends=DEPENDS)
-    vat_code = fields.Function(fields.Char('VAT Code'),
-        'get_vat_code', searcher='search_vat_code')
+        states=STATES, depends=DEPENDS,
+        help="Add other identifiers of the party.")
+    tax_identifier = fields.Function(fields.Many2One(
+            'party.identifier', 'Tax Identifier',
+            help="The identifier used for tax report."),
+        'get_tax_identifier', searcher='search_tax_identifier')
     addresses = fields.One2Many('party.address', 'party',
         'Addresses', states=STATES, depends=DEPENDS)
     contact_mechanisms = fields.One2Many('party.contact_mechanism', 'party',
         'Contact Mechanisms', states=STATES, depends=DEPENDS)
     categories = fields.Many2Many('party.party-party.category',
-        'party', 'category', 'Categories', states=STATES, depends=DEPENDS)
-    active = fields.Boolean('Active', select=True)
+        'party', 'category', 'Categories', states=STATES, depends=DEPENDS,
+        help="The categories the party belongs to.")
+    active = fields.Boolean('Active', select=True, states={
+            'readonly': Bool(Eval('replaced_by')),
+            },
+        depends=['replaced_by'],
+        help="Uncheck to exclude the party from future use.")
+    replaced_by = fields.Many2One('party.party', "Replaced By", readonly=True,
+        states={
+            'invisible': ~Eval('replaced_by'),
+            },
+        help="The party replacing this one.")
     full_name = fields.Function(fields.Char('Full Name'), 'get_full_name')
     phone = fields.Function(fields.Char('Phone'), 'get_mechanism')
     mobile = fields.Function(fields.Char('Mobile'), 'get_mechanism')
@@ -69,24 +88,20 @@ class Party(ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         pool = Pool()
-        Property = pool.get('ir.property')
+        PartyLang = pool.get('party.party.lang')
         TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
         table = cls.__table__()
+        party_lang = PartyLang.__table__()
 
         super(Party, cls).__register__(module_name)
 
         table_h = TableHandler(cls, module_name)
         if table_h.column_exist('lang'):
-            cursor.execute(*table.select(table.id, table.lang,
-                    order_by=table.lang))
-            for lang_id, group in groupby(cursor.fetchall(), lambda r: r[1]):
-                ids = [id_ for id_, _ in group]
-                if lang_id is not None:
-                    value = '%s,%s' % (cls.lang.model_name, lang_id)
-                else:
-                    value = None
-                Property.set('lang', cls.__name__, ids, value)
+            query = party_lang.insert(
+                [party_lang.party, party_lang.lang],
+                table.select(table.id, table.lang))
+            cursor.execute(*query)
             table_h.drop_column('lang')
 
         # Migration from 3.8
@@ -109,37 +124,55 @@ class Party(ModelSQL, ModelView):
     def default_addresses():
         if Transaction().user == 0:
             return []
-        Address = Pool().get('party.address')
-        fields_names = list(x for x in Address._fields.keys()
-            if x not in ('id', 'create_uid', 'create_date',
-                'write_uid', 'write_date'))
-        return [Address.default_get(fields_names)]
+        return [{}]
 
-    @staticmethod
-    def default_code_readonly():
+    @classmethod
+    def default_lang(cls, **pattern):
         Configuration = Pool().get('party.configuration')
         config = Configuration(1)
-        return bool(config.party_sequence)
+        lang = config.get_multivalue('party_lang', **pattern)
+        return lang.id if lang else None
+
+    @classmethod
+    def default_code_readonly(cls, **pattern):
+        Configuration = Pool().get('party.configuration')
+        config = Configuration(1)
+        return bool(config.get_multivalue('party_sequence', **pattern))
 
     def get_code_readonly(self, name):
         return True
 
     @classmethod
-    def _vat_types(cls):
+    def _tax_identifier_types(cls):
         return ['eu_vat']
 
-    def get_vat_code(self, name):
-        types = self._vat_types()
+    def get_tax_identifier(self, name):
+        types = self._tax_identifier_types()
         for identifier in self.identifiers:
             if identifier.type in types:
-                return identifier.code
+                return identifier.id
 
     @classmethod
-    def search_vat_code(cls, name, clause):
-        return [
-            ('identifiers.code',) + tuple(clause[1:]),
-            ('identifiers.type', 'in', cls._vat_types()),
+    def search_tax_identifier(cls, name, clause):
+        _, operator, value = clause
+        types = cls._tax_identifier_types()
+        domain = [
+            ('identifiers', 'where', [
+                    ('code', operator, value),
+                    ('type', 'in', types),
+                    ]),
             ]
+        # Add party without tax identifier
+        if ((operator == '=' and value is None)
+                or (operator == 'in' and None in value)):
+            domain = ['OR',
+                domain, [
+                    ('identifiers', 'not where', [
+                            ('type', 'in', types),
+                            ]),
+                    ],
+                ]
+        return domain
 
     def get_full_name(self, name):
         return self.name
@@ -151,15 +184,21 @@ class Party(ModelSQL, ModelView):
         return ''
 
     @classmethod
-    def create(cls, vlist):
-        Sequence = Pool().get('ir.sequence')
-        Configuration = Pool().get('party.configuration')
+    def _new_code(cls, **pattern):
+        pool = Pool()
+        Sequence = pool.get('ir.sequence')
+        Configuration = pool.get('party.configuration')
+        config = Configuration(1)
+        sequence = config.get_multivalue('party_sequence', **pattern)
+        if sequence:
+            return Sequence.get_id(sequence.id)
 
+    @classmethod
+    def create(cls, vlist):
         vlist = [x.copy() for x in vlist]
         for values in vlist:
             if not values.get('code'):
-                config = Configuration(1)
-                values['code'] = Sequence.get_id(config.party_sequence.id)
+                values['code'] = cls._new_code()
             values.setdefault('addresses', None)
         return super(Party, cls).create(vlist)
 
@@ -214,6 +253,32 @@ class Party(ModelSQL, ModelView):
         return default_address
 
 
+class PartyLang(ModelSQL, ValueMixin):
+    "Party Lang"
+    __name__ = 'party.party.lang'
+    party = fields.Many2One(
+        'party.party', "Party", ondelete='CASCADE', select=True)
+    lang = fields.Many2One('ir.lang', "Language")
+
+    @classmethod
+    def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
+        exist = TableHandler.table_exist(cls._table)
+
+        super(PartyLang, cls).__register__(module_name)
+
+        if not exist:
+            cls._migrate_property([], [], [])
+
+    @classmethod
+    def _migrate_property(cls, field_names, value_names, fields):
+        field_names.append('lang')
+        value_names.append('lang')
+        migrate_property(
+            'party.party', field_names, cls, value_names,
+            parent='party', fields=fields)
+
+
 class PartyCategory(ModelSQL):
     'Party - Category'
     __name__ = 'party.party-party.category'
@@ -224,13 +289,18 @@ class PartyCategory(ModelSQL):
         ondelete='CASCADE', required=True, select=True)
 
 
-class PartyIdentifier(ModelSQL, ModelView):
+class PartyIdentifier(sequence_ordered(), ModelSQL, ModelView):
     'Party Identifier'
     __name__ = 'party.identifier'
     _rec_name = 'code'
     party = fields.Many2One('party.party', 'Party', ondelete='CASCADE',
-        required=True, select=True)
-    type = fields.Selection('get_types', 'Type')
+        required=True, select=True,
+        help="The party identified by this record.")
+    type = fields.Selection([
+            (None, ''),
+            ('eu_vat', 'VAT'),
+            ], 'Type')
+    type_string = type.translated('type')
     code = fields.Char('Code', required=True)
 
     @classmethod
@@ -272,13 +342,6 @@ class PartyIdentifier(ModelSQL, ModelView):
             party_h.drop_column('vat_number')
             party_h.drop_column('vat_country')
 
-    @classmethod
-    def get_types(cls):
-        return [
-            (None, ''),
-            ('eu_vat', 'VAT'),
-            ]
-
     @fields.depends('type', 'code')
     def on_change_with_code(self):
         if self.type == 'eu_vat':
@@ -374,3 +437,123 @@ class CheckVIES(Wizard):
             'parties_succeed': [p.id for p in self.result.parties_succeed],
             'parties_failed': [p.id for p in self.result.parties_failed],
             }
+
+
+class PartyReplace(Wizard):
+    "Replace Party"
+    __name__ = 'party.replace'
+    start_state = 'ask'
+    ask = StateView('party.replace.ask', 'party.replace_ask_view_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Replace", 'replace', 'tryton-find-replace', default=True),
+            ])
+    replace = StateTransition()
+
+    @classmethod
+    def __setup__(cls):
+        super(PartyReplace, cls).__setup__()
+        cls._error_messages.update({
+                'different_name': ("Parties have different names: "
+                    "%(source_name)s vs %(destination_name)s."),
+                'different_tax_identifier': (
+                    "Parties have different Tax Identifier: "
+                    "%(source_code)s vs %(destination_code)s."),
+                })
+
+    def check_similarity(self):
+        source = self.ask.source
+        destination = self.ask.destination
+        if source.name != destination.name:
+            key = 'party.replace name %s %s' % (source.id, destination.id)
+            self.raise_user_warning(key, 'different_name', {
+                    'source_name': source.name,
+                    'destination_name': destination.name,
+                    })
+        source_code = (source.tax_identifier.code
+            if source.tax_identifier else '')
+        destination_code = (destination.tax_identifier.code
+            if destination.tax_identifier else '')
+        if source_code != destination_code:
+            key = 'party.replace tax_identifier %s %s' % (
+                source.id, destination.id)
+            self.raise_user_warning(key, 'different_tax_identifier', {
+                    'source_code': source_code,
+                    'destination_code': destination_code,
+                    })
+
+    def transition_replace(self):
+        pool = Pool()
+        Address = pool.get('party.address')
+        ContactMechanism = pool.get('party.contact_mechanism')
+        transaction = Transaction()
+
+        self.check_similarity()
+        source = self.ask.source
+        destination = self.ask.destination
+
+        Address.write(list(source.addresses), {
+                'active': False,
+                })
+        ContactMechanism.write(list(source.contact_mechanisms), {
+                'active': False,
+                })
+        source.replaced_by = destination
+        source.active = False
+        source.save()
+
+        cursor = transaction.connection.cursor()
+        for model_name, field_name in self.fields_to_replace():
+            Model = pool.get(model_name)
+            table = Model.__table__()
+            column = Column(table, field_name)
+            where = column == source.id
+
+            if transaction.database.has_returning():
+                returning = [table.id]
+            else:
+                cursor.execute(*table.select(table.id, where=where))
+                ids = [x[0] for x in cursor]
+                returning = None
+
+            cursor.execute(*table.update(
+                    [column],
+                    [destination.id],
+                    where=where,
+                    returning=returning))
+
+            if transaction.database.has_returning():
+                ids = [x[0] for x in cursor]
+
+            Model._insert_history(ids)
+        return 'end'
+
+    @classmethod
+    def fields_to_replace(cls):
+        return [
+            ('party.address', 'party'),
+            ('party.contact_mechanism', 'party'),
+            ]
+
+
+class PartyReplaceAsk(ModelView):
+    "Replace Party"
+    __name__ = 'party.replace.ask'
+    source = fields.Many2One('party.party', "Source", required=True,
+        help="The party to be replaced.")
+    destination = fields.Many2One('party.party', "Destination", required=True,
+        domain=[
+            ('id', '!=', Eval('source', -1)),
+            ],
+        depends=['source'],
+        help="The party that replaces.")
+
+    @classmethod
+    def default_source(cls):
+        context = Transaction().context
+        if context.get('active_model') == 'party.party':
+            return context.get('active_id')
+
+    @fields.depends('source')
+    def on_change_source(self):
+        if self.source and self.source.replaced_by:
+            self.destination = self.source.replaced_by
